import asyncio
import logging
from collections import deque
from typing import Optional

import numpy as np
from getstream.video.rtc.track_util import PcmData

logger = logging.getLogger(__name__)


class AudioQueue:
    """
    Queue for audio.
    - allows you to read a specific number of samples or duration of audio (needed for Silero)
    - supports a limit defined in seconds

    If the queue is full when adding to it, log a warning
    When using queue.get allow specifying either the number of samples or the duration

    """

    def __init__(self, buffer_limit_ms: int):
        """
        Initialize AudioQueue.

        Args:
            buffer_limit_ms: Maximum buffer duration in milliseconds
        """
        self.buffer_limit_ms = buffer_limit_ms
        self._buffer: deque[PcmData] = deque()
        self._total_samples = 0
        self._sample_rate: Optional[int] = None
        self._not_empty = asyncio.Event()
        self._lock = asyncio.Lock()

    def empty(self) -> bool:
        """Check if the queue is empty."""
        return len(self._buffer) == 0

    def qsize(self) -> int:
        """Get the number of items in the queue."""
        return len(self._buffer)

    def _current_duration_ms(self) -> float:
        """Get current buffer duration in milliseconds."""
        if self._sample_rate is None or self._total_samples == 0:
            return 0.0
        return (self._total_samples / self._sample_rate) * 1000

    async def put(self, item: PcmData) -> None:
        """
        Add PcmData to the queue.

        Args:
            item: PcmData to add to the queue

        Logs a warning if adding the item would exceed the buffer limit.
        """
        if not isinstance(item, PcmData):
            raise TypeError(f"AudioQueue only accepts PcmData, got {type(item)}")

        async with self._lock:
            # Track sample rate from first item
            if self._sample_rate is None:
                self._sample_rate = item.sample_rate
            elif self._sample_rate != item.sample_rate:
                logger.warning(
                    f"Sample rate mismatch: expected {self._sample_rate}, got {item.sample_rate}"
                )

            # Check if adding this would exceed the buffer limit
            new_samples = len(item.samples)
            new_total_samples = self._total_samples + new_samples
            new_duration_ms = (new_total_samples / item.sample_rate) * 1000

            if new_duration_ms > self.buffer_limit_ms:
                logger.warning(
                    f"AudioQueue buffer limit exceeded: {new_duration_ms:.1f}ms > {self.buffer_limit_ms}ms"
                )

            self._buffer.append(item)
            self._total_samples += new_samples
            self._not_empty.set()

    def put_nowait(self, item: PcmData) -> None:
        """
        Add PcmData to the queue without waiting.

        Args:
            item: PcmData to add to the queue

        Logs a warning if adding the item would exceed the buffer limit.
        """
        if not isinstance(item, PcmData):
            raise TypeError(f"AudioQueue only accepts PcmData, got {type(item)}")

        # Track sample rate from first item
        if self._sample_rate is None:
            self._sample_rate = item.sample_rate
        elif self._sample_rate != item.sample_rate:
            logger.warning(
                f"Sample rate mismatch: expected {self._sample_rate}, got {item.sample_rate}"
            )

        # Check if adding this would exceed the buffer limit
        new_samples = len(item.samples)
        new_total_samples = self._total_samples + new_samples
        new_duration_ms = (new_total_samples / item.sample_rate) * 1000

        if new_duration_ms > self.buffer_limit_ms:
            logger.warning(
                f"AudioQueue buffer limit exceeded: {new_duration_ms:.1f}ms > {self.buffer_limit_ms}ms"
            )

        self._buffer.append(item)
        self._total_samples += new_samples
        self._not_empty.set()

    async def get(self) -> PcmData:
        """
        Get the next PcmData from the queue.

        Returns:
            PcmData object from the queue
        """
        while True:
            async with self._lock:
                if self._buffer:
                    item = self._buffer.popleft()
                    self._total_samples -= len(item.samples)
                    if not self._buffer:
                        self._not_empty.clear()
                    return item

            # Wait for items to be added
            await self._not_empty.wait()

    def get_nowait(self) -> PcmData:
        """
        Get the next PcmData from the queue without waiting.

        Returns:
            PcmData object from the queue
        """
        if not self._buffer:
            raise asyncio.QueueEmpty("Queue is empty")

        item = self._buffer.popleft()
        self._total_samples -= len(item.samples)
        if not self._buffer:
            self._not_empty.clear()
        return item

    async def get_samples(self, num_samples: int, timeout: float = 0.1) -> PcmData:
        """
        Get a specific number of audio samples from the queue.

        Args:
            num_samples: Number of samples to retrieve
            timeout: Max time to wait for more samples when queue is empty (seconds)

        Returns:
            PcmData containing exactly num_samples (or less if queue empties)
        """
        collected_samples: list[np.ndarray] = []
        collected_count = 0
        last_item_format = None
        last_item_channels = 1
        last_participant = None

        while collected_count < num_samples:
            # Wait for items if queue is empty, with timeout
            if self.empty():
                # If we've already collected some samples, return what we have
                if collected_samples:
                    break
                # Otherwise wait for items
                try:
                    await asyncio.wait_for(self._not_empty.wait(), timeout=timeout)
                except asyncio.TimeoutError:
                    # No items arrived in time
                    if collected_samples:
                        break
                    raise asyncio.QueueEmpty("Queue is empty")

            async with self._lock:
                if not self._buffer:
                    continue

                item = self._buffer.popleft()
                self._total_samples -= len(item.samples)
                last_participant = item.participant

                last_item_format = item.format
                last_item_channels = item.channels

                samples_needed = num_samples - collected_count

                if len(item.samples) <= samples_needed:
                    # Use the entire item
                    collected_samples.append(item.samples)
                    collected_count += len(item.samples)
                else:
                    # Split the item - take what we need, put rest back
                    collected_samples.append(item.samples[:samples_needed])
                    collected_count += samples_needed

                    # Put the remainder back at the front
                    remainder = PcmData(
                        samples=item.samples[samples_needed:],
                        sample_rate=item.sample_rate,
                        format=item.format,
                        channels=item.channels,
                    )
                    self._buffer.appendleft(remainder)
                    self._total_samples += len(remainder.samples)
                    break

                # Clear event if buffer is now empty
                if not self._buffer:
                    self._not_empty.clear()

        if not collected_samples:
            raise asyncio.QueueEmpty("Queue is empty")

        # Concatenate all collected samples
        all_samples = np.concatenate(collected_samples)

        # Use properties from the last item we got
        pcm = PcmData(
            samples=all_samples,
            sample_rate=self._sample_rate,
            format=last_item_format,
            channels=last_item_channels,
        )
        pcm.participant = last_participant
        return pcm

    async def get_duration(self, duration_ms: float) -> PcmData:
        """
        Get audio data for a specific duration from the queue.

        Args:
            duration_ms: Duration in milliseconds to retrieve

        Returns:
            PcmData containing audio for the requested duration (or less if queue empties)
        """
        # Wait for first item if needed to determine sample rate
        if self._sample_rate is None:
            await self._not_empty.wait()

        if self._sample_rate is None:
            raise asyncio.QueueEmpty("Queue is empty")

        # Convert duration to number of samples
        num_samples = int((duration_ms / 1000) * self._sample_rate)
        return await self.get_samples(num_samples)

    def get_buffer_info(self) -> dict:
        """
        Get information about the current buffer state.

        Returns:
            Dict with buffer statistics
        """
        return {
            "buffer_limit_ms": self.buffer_limit_ms,
            "current_duration_ms": self._current_duration_ms(),
            "total_samples": self._total_samples,
            "sample_rate": self._sample_rate,
            "num_chunks": len(self._buffer),
            "queue_size": self.qsize(),
        }



"""
Generic CLI runner for Vision Agents examples.

Provides a Click-based CLI with common options for debugging and logging.
"""

import asyncio
import logging
import warnings
from typing import TYPE_CHECKING, Optional
from uuid import uuid4

import click
from vision_agents.core.utils.logging import configure_sdk_logger

if TYPE_CHECKING:
    from vision_agents.core.agents.agent_launcher import AgentLauncher


asyncio_logger = logging.getLogger("asyncio")

logger = logging.getLogger(__name__)


def cli(launcher: "AgentLauncher") -> None:
    """
    Create and run a CLI from an AgentLauncher.

    Usage:
        if __name__ == "__main__":
            cli(AgentLauncher(create_agent=create_agent, join_call=join_call))

    Args:
        launcher: AgentLauncher instance with create_agent and join_call functions
    """

    @click.command()
    @click.option(
        "--call-type",
        type=str,
        default="default",
        help="Call type for the video call",
    )
    @click.option(
        "--call-id",
        type=str,
        default=None,
        help="Call ID for the video call (auto-generated if not provided)",
    )
    @click.option(
        "--debug",
        is_flag=True,
        default=False,
        help="Enable debug mode",
    )
    @click.option(
        "--log-level",
        type=click.Choice(
            ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
        ),
        default="INFO",
        help="Set the logging level",
    )
    @click.option(
        "--no-demo",
        is_flag=True,
        default=False,
        help="Disable opening the demo UI",
    )
    @click.option(
        "--video-track-override",
        type=click.Path(dir_okay=False, exists=True, resolve_path=True),
        default=None,
        help="Optional local video track override for debugging. "
        "This track will play instead of any incoming video track.",
    )
    def run_agent(
        call_type: str,
        call_id: Optional[str],
        debug: bool,
        log_level: str,
        no_demo: bool,
        video_track_override: Optional[str],
    ) -> None:
        """Run the agent with the specified configuration."""
        # Configure logging
        numeric_level = getattr(logging, log_level.upper(), logging.INFO)
        configure_sdk_logger(level=numeric_level)

        # Suppress dataclasses_json missing value RuntimeWarnings.
        # They pollute the output and cannot be fixed by the users.
        warnings.filterwarnings(
            "ignore", category=RuntimeWarning, module="dataclasses_json.core"
        )

        # Generate call ID if not provided
        if call_id is None:
            call_id = str(uuid4())

        async def _run():
            logger.info("🚀 Launching agent...")

            try:
                # Start the agent launcher.
                await launcher.start()

                # Create the agent
                agent = await launcher.launch()
                if video_track_override:
                    agent.set_video_track_override_path(video_track_override)

                logger.info("✅ Agent warmed up and ready")

                # Open demo UI by default
                if (
                    not no_demo
                    and hasattr(agent, "edge")
                    and hasattr(agent.edge, "open_demo_for_agent")
                ):
                    logger.info("🌐 Opening demo UI...")
                    await agent.edge.open_demo_for_agent(agent, call_type, call_id)

                # Join call if join_call function is provided
                if launcher.join_call:
                    logger.info(f"📞 Joining call: {call_type}/{call_id}")
                    result = launcher.join_call(agent, call_type, call_id)
                    if asyncio.iscoroutine(result):
                        await result
                else:
                    logger.warning(
                        '⚠️ No "join_call" function provided; the agent is created but will not join the call'
                    )
            except KeyboardInterrupt:
                logger.info("🛑 Received interrupt signal, shutting down gracefully...")
            except Exception as e:
                logger.error(f"❌ Error running agent: {e}", exc_info=True)
                raise
            finally:
                await launcher.stop()

        asyncio_logger_level = asyncio_logger.level

        try:
            asyncio.run(_run(), debug=debug)
        except KeyboardInterrupt:
            # Temporarily suppress asyncio error logging during cleanup
            asyncio_logger_level = asyncio_logger.level
            # Suppress KeyboardInterrupt and asyncio errors during cleanup
            asyncio_logger.setLevel(logging.CRITICAL)
            logger.info("👋 Agent shutdown complete")
        finally:
            # Restore original logging level
            asyncio_logger.setLevel(asyncio_logger_level)

    # Invoke the click command
    run_agent()



from typing import List

from vision_agents.core.stt.events import STTTranscriptEvent, STTPartialTranscriptEvent


class TranscriptBuffer:
    """
    Buffer for accumulating transcript text from STT events.

    Partial events update the current working segment. Final events
    finalize the segment and any subsequent events start a new one.

    Example flow:
    - Partial "I" → ["I"]
    - Partial "I am" → ["I am"]
    - Final "I am walking" → ["I am walking"]
    - Partial "To" → ["I am walking", "To"]
    - Final "To the store" → ["I am walking", "To the store"]

    It's also possible you receive corrections
    - Partial "It's damp"
    - Partial "It's a swamp"
    - Final "It's a swamp"
    """

    def __init__(self):
        self._segments: List[str] = []
        self._has_pending_partial: bool = False

    def update(
        self, event: STTTranscriptEvent | STTPartialTranscriptEvent | str
    ) -> None:
        """
        Update the buffer from an STT event or text string.

        Args:
            event: Either an STT event or a plain text string.

        Partial events update the current working segment.
        Final events (STTTranscriptEvent or strings) finalize the segment.
        """
        text = event if isinstance(event, str) else event.text
        text = text.strip()
        if not text:
            return

        is_partial = isinstance(event, STTPartialTranscriptEvent)

        if is_partial:
            if self._has_pending_partial and self._segments:
                # Update the existing partial segment (unless text matches)
                if self._segments[-1] != text:
                    self._segments[-1] = text
            else:
                # Start a new partial segment only if different from last finalized
                if not self._segments or self._segments[-1] != text:
                    self._segments.append(text)
                    self._has_pending_partial = True
        else:
            # Final event (STTTranscriptEvent or string)
            if self._has_pending_partial and self._segments:
                # Replace the partial with the final text
                self._segments[-1] = text
            else:
                # No pending partial - add as new segment only if different from last
                if not self._segments or self._segments[-1] != text:
                    self._segments.append(text)
            self._has_pending_partial = False

    def reset(self) -> None:
        """Clear all accumulated segments."""
        self._segments.clear()
        self._has_pending_partial = False

    @property
    def segments(self) -> List[str]:
        """Return a copy of the current segments."""
        return self._segments.copy()

    @property
    def text(self) -> str:
        """Return all segments joined with spaces."""
        return " ".join(self._segments)

    def __len__(self) -> int:
        return len(self._segments)

    def __bool__(self) -> bool:
        return bool(self._segments)



import asyncio
import contextlib
import importlib.metadata
import inspect
import logging
import os
from typing import Awaitable, Callable, Dict, Optional, ParamSpec, TypeVar, Union

import httpx

logger = logging.getLogger(__name__)


T = TypeVar("T")
P = ParamSpec("P")


# Type alias for markdown file contents: maps filename to file content
MarkdownFileContents = Dict[str, str]

# Cache version at module load time to avoid blocking I/O during async operations
_VISION_AGENTS_VERSION: str | None = None


def _load_version() -> str:
    """Load version once at module import time."""
    try:
        return importlib.metadata.version("vision-agents")
    except importlib.metadata.PackageNotFoundError:
        return "unknown"


_VISION_AGENTS_VERSION = _load_version()


def get_vision_agents_version() -> Optional[str]:
    """
    Get the installed vision-agents package version.

    Returns:
        Version string, or "unknown" if not available.
    """
    return _VISION_AGENTS_VERSION or "unknown"


async def ensure_model(path: str, url: str) -> str:
    """
    Download a model file asynchronously if it doesn't exist.

    Args:
        path: Local path where the model should be saved
        url: URL to download the model from

    Returns:
        The path to the model file
    """

    if not os.path.exists(path):
        model_name = os.path.basename(path)
        logger.info(f"Downloading {model_name}...")

        try:
            async with httpx.AsyncClient(
                timeout=300.0, follow_redirects=True
            ) as client:
                async with client.stream("GET", url) as response:
                    response.raise_for_status()

                    # Write file in chunks to avoid loading entire file in memory
                    chunks = []
                    async for chunk in response.aiter_bytes(chunk_size=8192):
                        chunks.append(chunk)

                    # Write all chunks to file in thread to avoid blocking event loop
                    def write_file():
                        with open(path, "wb") as f:
                            for chunk in chunks:
                                f.write(chunk)

                    await asyncio.to_thread(write_file)

            logger.info(f"{model_name} downloaded.")
        except httpx.HTTPError as e:
            # Clean up partial download on error
            if os.path.exists(path):
                os.remove(path)
            raise RuntimeError(f"Failed to download {model_name}: {e}")

    return path


async def await_or_run(
    fn: Callable[P, Union[T, Awaitable[T]]], *args: P.args, **kwargs: P.kwargs
) -> T:
    """
    Run a callable that may be synchronous or asynchronous.
    Automatically awaits if the result is awaitable.
    """
    result = fn(*args, **kwargs)
    if inspect.isawaitable(result):
        return await result
    return result


async def cancel_and_wait(fut: asyncio.Future) -> None:
    """
    Cancel an async task or future and wait for it to complete.

    Args:
        fut: a Future or Task to cancel.

    Returns:
        None
    """

    if fut.done():
        return None
    fut.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await asyncio.shield(fut)
    return None



"""Utility to convert MCP tools to function registry format."""

from typing import Any, Dict, Callable
from mcp import types

from ..llm.llm_types import ToolSchema


class MCPToolConverter:
    """Converts MCP tools to function registry format."""

    @staticmethod
    def mcp_tool_to_tool_schema(tool: types.Tool) -> ToolSchema:
        """Convert an MCP tool to a ToolSchema.

        Args:
            tool: MCP tool object

        Returns:
            ToolSchema compatible with function registry
        """
        # Convert MCP tool input schema to JSON schema format
        parameters_schema = MCPToolConverter._convert_input_schema(tool.inputSchema)

        return ToolSchema(
            name=tool.name,
            description=tool.description or "",
            parameters_schema=parameters_schema,
        )

    @staticmethod
    def _convert_input_schema(input_schema: Dict[str, Any]) -> Dict[str, Any]:
        """Convert MCP input schema to JSON schema format.

        Args:
            input_schema: MCP tool input schema

        Returns:
            JSON schema compatible with function registry
        """
        # MCP tools already use JSON schema format, so we can mostly pass through
        # but we need to ensure it has the right structure
        schema = input_schema.copy()

        # Ensure required fields are present
        if "type" not in schema:
            schema["type"] = "object"

        if "properties" not in schema:
            schema["properties"] = {}

        # Ensure additionalProperties is set
        if "additionalProperties" not in schema:
            schema["additionalProperties"] = False

        return schema

    @staticmethod
    def create_mcp_tool_wrapper(
        server_index: int, tool_name: str, agent_ref
    ) -> "Callable":
        """Create a wrapper function for calling MCP tools.

        Args:
            server_index: Index of the MCP server in the agent's mcp_servers list
            tool_name: Name of the tool to call
            agent_ref: Reference to the agent instance

        Returns:
            Callable function that can be registered with the function registry
        """

        async def mcp_tool_wrapper(**kwargs) -> Any:
            """Wrapper function for MCP tool calls."""
            try:
                result = await agent_ref.call_tool(server_index, tool_name, kwargs)
                # Extract the actual result from MCP response
                if hasattr(result, "content") and result.content:
                    # MCP tools return CallToolResult with content
                    if isinstance(result.content, list) and len(result.content) > 0:
                        # Get the first content item
                        content_item = result.content[0]
                        if hasattr(content_item, "text"):
                            return content_item.text
                        elif hasattr(content_item, "data"):
                            return content_item.data
                        else:
                            return str(content_item)
                    else:
                        return str(result.content)
                else:
                    return str(result)
            except Exception as e:
                # Return error information in a structured way
                return {
                    "error": str(e),
                    "tool": tool_name,
                    "server_index": server_index,
                }

        return mcp_tool_wrapper



from __future__ import annotations

import abc
import asyncio
import json
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
)

import aiortc
from vision_agents.core.instructions import Instructions
from vision_agents.core.llm import events
from vision_agents.core.llm.events import ToolEndEvent, ToolStartEvent

if TYPE_CHECKING:
    from vision_agents.core.agents import Agent
    from vision_agents.core.agents.conversation import Conversation

from getstream.video.rtc import PcmData
from getstream.video.rtc.pb.stream.video.sfu.models.models_pb2 import Participant
from vision_agents.core.events.manager import EventManager
from vision_agents.core.processors import Processor

from ..utils.video_forwarder import VideoForwarder
from .function_registry import FunctionRegistry
from .llm_types import NormalizedToolCallItem, ToolSchema

T = TypeVar("T")


class LLMResponseEvent(Generic[T]):
    def __init__(self, original: T, text: str, exception: Optional[Exception] = None):
        self.original = original
        self.text = text
        self.exception = exception


BeforeCb = Callable[[List[Any]], None]
AfterCb = Callable[[LLMResponseEvent], None]


class LLM(abc.ABC):
    before_response_listener: BeforeCb
    after_response_listener: AfterCb
    agent: Optional["Agent"]
    function_registry: FunctionRegistry

    def __init__(self):
        super().__init__()
        self.agent = None
        self.events = EventManager()
        self.events.register_events_from_module(events)
        self.function_registry = FunctionRegistry()
        # LLM instructions. Provided by the Agent via `set_instructions` method
        self._instructions: str = ""
        self._conversation: Optional[Conversation] = None

    async def simple_response(
        self,
        text: str,
        processors: Optional[List[Processor]] = None,
        participant: Optional[Participant] = None,
    ) -> LLMResponseEvent[Any]:
        raise NotImplementedError

    def _get_tools_for_provider(self) -> List[Dict[str, Any]]:
        """
        Get tools in provider-specific format.
        This method should be overridden by each LLM implementation.

        Returns:
            List of tools in the provider's expected format.
        """
        tools = self.get_available_functions()
        return self._convert_tools_to_provider_format(tools)

    def _convert_tools_to_provider_format(
        self, tools: List[ToolSchema]
    ) -> List[Dict[str, Any]]:
        """
        Convert ToolSchema objects to provider-specific format.
        This method should be overridden by each LLM implementation.

        Args:
            tools: List of ToolSchema objects

        Returns:
            List of tools in provider-specific format
        """
        # Default implementation - should be overridden
        return []

    def _extract_tool_calls_from_response(
        self, response: Any
    ) -> List[NormalizedToolCallItem]:
        """
        Extract tool calls from provider-specific response.
        This method should be overridden by each LLM implementation.

        Args:
            response: Provider-specific response object

        Returns:
            List of normalized tool call items
        """
        # Default implementation - should be overridden
        return []

    def _extract_tool_calls_from_stream_chunk(
        self, chunk: Any
    ) -> List[NormalizedToolCallItem]:
        """
        Extract tool calls from a streaming chunk.
        This method should be overridden by each LLM implementation.

        Args:
            chunk: Provider-specific streaming chunk

        Returns:
            List of normalized tool call items
        """
        # Default implementation - should be overridden
        return []

    def _create_tool_result_message(
        self, tool_calls: List[NormalizedToolCallItem], results: List[Any]
    ) -> List[Dict[str, Any]]:
        """
        Create tool result messages for the provider.
        This method should be overridden by each LLM implementation.

        Args:
            tool_calls: List of tool calls that were executed
            results: List of results from function execution

        Returns:
            List of tool result messages in provider format
        """
        # Default implementation - should be overridden
        return []

    def _attach_agent(self, agent: Agent):
        """
        Attach agent to the llm
        """
        self.agent = agent
        self.set_instructions(agent.instructions)

    def set_conversation(self, conversation: Conversation):
        """
        Provide the Conversation object to the LLM to access the chat history.
        To be called by the Agent after it joins the call.

        Args:
            conversation: a Conversation object

        Returns:
        """
        self._conversation = conversation

    def set_instructions(self, instructions: Instructions | str) -> None:
        """
        Set instructions for LLM.

        Args:
            instructions: instructions object. Can be either `str` or `Instructions`.
        """
        if isinstance(instructions, str):
            self._instructions = instructions
        elif isinstance(instructions, Instructions):
            self._instructions = instructions.full_reference
        else:
            raise TypeError(
                f"Invalid instructions type {type(instructions)}, expected str or Instructions"
            )

    def register_function(
        self, name: Optional[str] = None, description: Optional[str] = None
    ) -> Callable:
        """
        Decorator to register a function with the LLM's function registry.

        Args:
            name: Optional custom name for the function. If not provided, uses the function name.
            description: Optional description for the function. If not provided, uses the docstring.

        Returns:
            Decorator function.
        """
        return self.function_registry.register(name, description)

    def get_available_functions(self) -> List[ToolSchema]:
        """Get a list of available function schemas."""
        return self.function_registry.get_tool_schemas()

    def call_function(self, name: str, arguments: Dict[str, Any]) -> Any:
        """
        Call a registered function with the given arguments.

        Args:
            name: Name of the function to call.
            arguments: Dictionary of arguments to pass to the function.

        Returns:
            Result of the function call.
        """
        return self.function_registry.call_function(name, arguments)

    def _tc_key(self, tc: NormalizedToolCallItem) -> Tuple[Optional[str], str, str]:
        """Generate a unique key for tool call deduplication.

        Args:
            tc: Tool call dictionary

        Returns:
            Tuple of (id, name, arguments_json) for deduplication
        """
        return (
            tc.get("id"),
            tc["name"],
            json.dumps(tc.get("arguments_json", {}), sort_keys=True),
        )

    async def _maybe_await(self, x):
        """Await if x is a coroutine, otherwise return x directly.

        Args:
            x: Value that might be a coroutine

        Returns:
            Awaited result if coroutine, otherwise x
        """
        if asyncio.iscoroutine(x):
            return await x
        return x

    async def _run_one_tool(self, tc: Dict[str, Any], timeout_s: float):
        """Run a single tool call with timeout.

        Args:
            tc: Tool call dictionary
            timeout_s: Timeout in seconds

        Returns:
            Tuple of (tool_call, result, error)
        """
        import inspect
        import time

        args = tc.get("arguments_json", tc.get("arguments", {})) or {}
        start_time = time.time()

        async def _invoke():
            # Get the actual function to check if it's async
            if hasattr(self.function_registry, "get_callable"):
                fn = self.function_registry.get_callable(tc["name"])
                if inspect.iscoroutinefunction(fn):
                    return await fn(**args)
                else:
                    # Run sync function in a worker thread to avoid blocking
                    return await asyncio.to_thread(fn, **args)
            else:
                # Fallback to existing call_function method
                res = self.call_function(tc["name"], args)
                return await self._maybe_await(res)

        try:
            # Send tool start event
            self.events.send(
                ToolStartEvent(
                    plugin_name="llm",
                    tool_name=tc["name"],
                    arguments=args,
                    tool_call_id=tc.get("id"),
                )
            )

            res = await asyncio.wait_for(_invoke(), timeout=timeout_s)
            execution_time = (time.time() - start_time) * 1000

            # Send tool end event (success)
            self.events.send(
                ToolEndEvent(
                    plugin_name="llm",
                    tool_name=tc["name"],
                    success=True,
                    result=res,
                    tool_call_id=tc.get("id"),
                    execution_time_ms=execution_time,
                )
            )

            return tc, res, None
        except Exception as e:
            execution_time = (time.time() - start_time) * 1000

            # Send tool end event (error)
            self.events.send(
                ToolEndEvent(
                    plugin_name="llm",
                    tool_name=tc["name"],
                    success=False,
                    error=str(e),
                    tool_call_id=tc.get("id"),
                    execution_time_ms=execution_time,
                )
            )

            return tc, {"error": str(e)}, e

    async def _execute_tools(
        self,
        calls: List[NormalizedToolCallItem],
        *,
        max_concurrency: int = 8,
        timeout_s: float = 30,
    ):
        """Execute multiple tool calls concurrently with timeout.

        Args:
            calls: List of tool call dictionaries
            max_concurrency: Maximum number of concurrent tool executions
            timeout_s: Timeout per tool execution in seconds

        Returns:
            List of tuples (tool_call, result, error)
        """
        sem = asyncio.Semaphore(max_concurrency)

        async def _guarded(tc):
            async with sem:
                return await self._run_one_tool(tc, timeout_s)

        return await asyncio.gather(*[_guarded(tc) for tc in calls])

    async def _dedup_and_execute(
        self,
        calls: List[NormalizedToolCallItem],
        *,
        max_concurrency: int = 8,
        timeout_s: float = 30,
        seen: Optional[set] = None,
    ):
        """De-duplicate (by id/name/args) then execute concurrently.

        Args:
            calls: List of tool call dictionaries
            max_concurrency: Maximum number of concurrent tool executions
            timeout_s: Timeout per tool execution in seconds
            seen: Set of seen tool call keys for deduplication

        Returns:
            Tuple of (triples, updated_seen_set)
        """
        seen = seen or set()
        to_run: List[NormalizedToolCallItem] = []
        for tc in calls:
            key = self._tc_key(tc)
            if key in seen:
                continue
            seen.add(key)
            to_run.append(tc)

        if not to_run:
            return [], seen  # nothing new

        triples = await self._execute_tools(
            to_run, max_concurrency=max_concurrency, timeout_s=timeout_s
        )
        return triples, seen

    def _sanitize_tool_output(self, value: Any, max_chars: int = 60_000) -> str:
        """Sanitize tool output to prevent oversized responses.

        Args:
            value: Tool output value (can be string, dict, or exception)
            max_chars: Maximum characters allowed

        Returns:
            Sanitized string output
        """
        if isinstance(value, str):
            s = value
        elif isinstance(value, Exception):
            s = f"Error: {type(value).__name__}: {value}"
        else:
            s = json.dumps(value)
        return (s[:max_chars] + "…") if len(s) > max_chars else s


class AudioLLM(LLM, metaclass=abc.ABCMeta):
    """
    A base class for LLMs capable of processing speech-to-speech audio.
    These models do not require TTS and STT services to run.
    """

    @abc.abstractmethod
    async def simple_audio_response(
        self, pcm: PcmData, participant: Optional[Participant] = None
    ):
        """
        Implement this method to forward PCM audio frames to the LLM.

        The audio should be raw PCM matching the model's expected
        format (typically 48 kHz mono, 16-bit).

        Args:
            pcm: PCM audio frame to forward upstream.
            participant: Optional participant information for the audio source.
        """


class VideoLLM(LLM, metaclass=abc.ABCMeta):
    """
    A base class for LLMs capable of processing video.

    These models will receive the video track from the `Agent` to analyze it.
    """

    @abc.abstractmethod
    async def watch_video_track(
        self,
        track: aiortc.mediastreams.MediaStreamTrack,
        shared_forwarder: Optional[VideoForwarder] = None,
    ) -> None:
        """
        Implement this method to watch and forward video tracks.

        Args:
            track: Video track to watch and forward.
            shared_forwarder: Optional shared VideoForwarder instance to use instead
                of creating a new one. Allows multiple consumers to share the same
                video stream.
        """

    @abc.abstractmethod
    async def stop_watching_video_track(self) -> None:
        """Stop watching the video track."""
        pass


class OmniLLM(AudioLLM, VideoLLM, metaclass=abc.ABCMeta):
    """
    A base class for LLMs capable of both video and speech-to-speech audio processing.
    """

    ...



import logging
import os
import re
from pathlib import Path

__all__ = ["Instructions", "InstructionsReadError"]


logger = logging.getLogger(__name__)

_INITIAL_CWD = os.getcwd()

_MD_PATTERN = re.compile(r"@([^\s@]+)")


class InstructionsReadError(Exception): ...


class Instructions:
    """
    Container for parsed instructions with input text and markdown files.

    Attributes:
        input_text: Input text that may contain @ mentioned markdown files.
        full_reference: Full reference that includes input text and contents of @ mentioned markdown files.
    """

    def __init__(self, input_text: str = "", base_dir: str | Path = ""):
        """
        Initialize Instructions object.

        Args:
            input_text: Input text that may contain @ mentioned markdown files.
                Ignores files starting with ".", non-md files, and files outside the base directory.

            base_dir: Base directory to search for markdown files. Default - current working directory.
        """
        self._base_dir = Path(base_dir or _INITIAL_CWD).resolve()
        self.input_text = input_text
        self.full_reference = self._extract_full_reference()

    def _extract_full_reference(self) -> str:
        """
        Parse instructions from an input text string, extracting @ mentioned markdown files and their contents.

        Returns:
            Instructions object containing the input text and file contents
        """

        # Find all @ mentions that look like markdown files
        matches = _MD_PATTERN.findall(self.input_text)

        # Create a dictionary mapping filename to file content
        markdown_contents = {}

        markdown_lines = [self.input_text]
        # Iterate over found @ mentions and try reading instructions from them
        for match in matches:
            # Try to read the markdown file content
            content = self._read_md_file(file_path=match)
            markdown_contents[match] = content

        # Add markdown file contents if any exist
        if markdown_contents:
            markdown_lines.append("\n\n## Referenced Documentation:")
            for filename, content in markdown_contents.items():
                markdown_lines.append(f"\n### {filename}")
                # Only include non-empty content
                markdown_lines.append(content or "*(File is empty)*")
        full_reference = "\n".join(markdown_lines)
        return full_reference

    def _read_md_file(self, file_path: str | Path) -> str:
        """
        Synchronous helper to read a markdown file.

        Args:
            file_path: Absolute or relative path to markdown file.
                Paths outside the base directory are ignored.
        """
        # Resolve the markdown file path
        file_path = Path(file_path)
        full_path = (
            file_path.resolve()
            if file_path.is_absolute()
            else (self._base_dir / file_path).resolve()
        )

        # Check if the path is a file, it exists, and it's a markdown file.
        skip_reason = ""
        if not full_path.exists():
            skip_reason = "file not found"
        elif not full_path.is_file():
            skip_reason = "path is not a file"
        elif full_path.name.startswith("."):
            skip_reason = 'filename cannot start with "."'
        elif full_path.suffix != ".md":
            skip_reason = "file is not .md"
        # The markdown file also must be inside the base_dir
        elif not full_path.is_relative_to(self._base_dir):
            skip_reason = f"path outside the base directory {self._base_dir}"

        if skip_reason:
            raise InstructionsReadError(
                f"Failed to read instructions from {full_path}; reason - {skip_reason}"
            )

        try:
            logger.info(f"Reading instructions from file {full_path}")
            with open(full_path, mode="r") as f:
                return f.read()
        except (OSError, IOError, UnicodeDecodeError) as exc:
            raise InstructionsReadError(
                f"Failed to read instructions from file {full_path}; reason - {exc}"
            ) from exc



"""Stream Agents Observability Package.

This package provides observability tools including metrics collection for Stream Agents.

Usage:
    # Configure OpenTelemetry in your application first, then:
    from vision_agents.core.observability import MetricsCollector

    agent = Agent(llm=OpenAILLM(), stt=DeepgramSTT(), tts=CartesiaTTS())
    collector = MetricsCollector(agent)  # Opt-in to metrics collection
"""

from .agent import AgentMetrics
from .collector import MetricsCollector
from .metrics import (
    llm_errors,
    llm_input_tokens,
    # LLM metrics
    llm_latency_ms,
    llm_output_tokens,
    llm_time_to_first_token_ms,
    llm_tool_calls,
    llm_tool_latency_ms,
    meter,
    realtime_agent_transcriptions,
    realtime_audio_input_bytes,
    realtime_audio_input_duration_ms,
    realtime_audio_output_bytes,
    realtime_audio_output_duration_ms,
    realtime_errors,
    realtime_responses,
    realtime_session_duration_ms,
    # Realtime LLM metrics
    realtime_sessions,
    realtime_user_transcriptions,
    stt_audio_duration_ms,
    stt_errors,
    # STT metrics
    stt_latency_ms,
    tts_audio_duration_ms,
    tts_characters,
    tts_errors,
    # TTS metrics
    tts_latency_ms,
    # Turn detection metrics
    turn_duration_ms,
    turn_trailing_silence_ms,
    video_detections,
    # Video processor metrics
    video_frames_processed,
    video_processing_latency_ms,
    vlm_errors,
    # VLM metrics
    vlm_inference_latency_ms,
    vlm_inferences,
    vlm_input_tokens,
    vlm_output_tokens,
)

__all__ = [
    # Main class
    "MetricsCollector",
    # Meter
    "meter",
    # STT metrics
    "stt_latency_ms",
    "stt_audio_duration_ms",
    "stt_errors",
    # TTS metrics
    "tts_latency_ms",
    "tts_audio_duration_ms",
    "tts_characters",
    "tts_errors",
    # LLM metrics
    "llm_latency_ms",
    "llm_time_to_first_token_ms",
    "llm_input_tokens",
    "llm_output_tokens",
    "llm_errors",
    "llm_tool_calls",
    "llm_tool_latency_ms",
    # Turn detection metrics
    "turn_duration_ms",
    "turn_trailing_silence_ms",
    # Realtime LLM metrics
    "realtime_sessions",
    "realtime_session_duration_ms",
    "realtime_audio_input_bytes",
    "realtime_audio_output_bytes",
    "realtime_audio_input_duration_ms",
    "realtime_audio_output_duration_ms",
    "realtime_responses",
    "realtime_user_transcriptions",
    "realtime_agent_transcriptions",
    "realtime_errors",
    # VLM metrics
    "vlm_inference_latency_ms",
    "vlm_inferences",
    "vlm_input_tokens",
    "vlm_output_tokens",
    "vlm_errors",
    # Video processor metrics
    "video_frames_processed",
    "video_processing_latency_ms",
    "video_detections",
    "AgentMetrics",
]



import asyncio
import os
import time
from typing_extensions import Self

import numpy as np
import onnxruntime as ort
from getstream.video.rtc import PcmData
from vision_agents.core.utils.utils import ensure_model

__all__ = ("SileroVADSession", "SileroVADSessionPool", "SILERO_CHUNK")

SILERO_CHUNK = 512
SILERO_ONNX_FILENAME = "silero_vad.onnx"
SILERO_ONNX_URL = "https://github.com/snakers4/silero-vad/raw/master/src/silero_vad/data/silero_vad.onnx"


class SileroVADSessionPool:
    """
    Initialize Silero VAD pool.
    It can be used to spawn stateful `SileroVAD` re-using the same inference session object.

    Args:
        model_path: Path to the ONNX model file
    """

    def __init__(self, model_path: str):
        opts = ort.SessionOptions()
        opts.inter_op_num_threads = 1
        self._model = ort.InferenceSession(model_path, sess_options=opts)

    @classmethod
    async def load(cls, model_dir: str) -> Self:
        """
        Initialize Silero VAD pool asynchronously without blocking event loop.

        Args:
            model_dir: path to the model directory.

        Returns:
            an instance of SileroVADSessionPool
        """
        path = os.path.join(model_dir, SILERO_ONNX_FILENAME)
        await ensure_model(path, SILERO_ONNX_URL)
        # Initialize VAD in thread pool to avoid blocking event loop
        pool = await asyncio.to_thread(  # type: ignore[func-returns-value]
            lambda: cls(  # type: ignore[arg-type]
                path
            )
        )
        return pool

    def session(self, reset_interval_seconds: float = 5.0) -> "SileroVADSession":
        """
        Create a new Silero VAD session.
        Each session has its own state but re-uses the same onnx inference session.

        Args:
            reset_interval_seconds: Reset internal state every N seconds to prevent drift

        Returns:
            an instace of `SileroVADSession`
        """
        return SileroVADSession(
            model=self._model, reset_interval_seconds=reset_interval_seconds
        )


class SileroVADSession:
    """
    Minimal Silero VAD ONNX wrapper
    """

    def __init__(self, model: ort.InferenceSession, reset_interval_seconds: float):
        """
        Initialize Silero VAD session.
        Each session has its own state but re-uses the same onnx inference session.

        Args:
            reset_interval_seconds: Reset internal state every N seconds to prevent drift
        """

        self._model = model
        self._context_size = 64  # Silero uses 64-sample context at 16 kHz
        self._reset_interval_seconds = reset_interval_seconds
        self._state: np.ndarray = np.zeros((2, 1, 128), dtype=np.float32)  # (2, B, 128)
        self._context: np.ndarray = np.zeros((1, 64), dtype=np.float32)
        self._init_states()

    def predict_speech(self, pcm: PcmData):
        # convert from pcm to the right format for silero

        chunks = pcm.resample(16000, 1).to_float32().chunks(SILERO_CHUNK, pad_last=True)
        scores = [self._predict_speech(c.samples) for c in chunks]
        return max(scores)

    def _init_states(self):
        self._state = np.zeros((2, 1, 128), dtype=np.float32)  # (2, B, 128)
        self._context = np.zeros((1, self._context_size), dtype=np.float32)
        self._last_reset_time = time.time()

    def _maybe_reset(self):
        if (time.time() - self._last_reset_time) >= self._reset_interval_seconds:
            self._init_states()

    def _predict_speech(self, chunk_f32: np.ndarray) -> float:
        """
        Compute speech probability for one chunk of length 512 (float32, mono).
        Returns a scalar float.
        """
        # Ensure shape (1, 512) and concat context
        x = np.reshape(chunk_f32, (1, -1))
        if x.shape[1] != SILERO_CHUNK:
            # Raise on incorrect usage
            raise ValueError(
                f"incorrect usage for predict speech. only send audio data in chunks of 512. got {x.shape[1]}"
            )
        x = np.concatenate((self._context, x), axis=1)

        # Run ONNX
        ort_inputs = {
            "input": x.astype(np.float32),
            "state": self._state,
            "sr": np.array(16000, dtype=np.int64),
        }
        outputs = self._model.run(None, ort_inputs)
        out, self._state = outputs

        # Update context (keep last 64 samples)
        self._context = x[:, -self._context_size :]
        self._maybe_reset()

        # out shape is (1, 1) -> return scalar
        prediction = float(out[0][0])
        return prediction



from enum import Enum
from typing import Any, Dict, List, Literal, TypedDict, Union

from typing_extensions import NotRequired


class TextPart(TypedDict):
    type: Literal["text"]
    text: str


class ImageBytesPart(TypedDict):
    type: Literal["image"]
    data: bytes
    mime_type: NotRequired[str]


class ImageURLPart(TypedDict):
    type: Literal["image"]
    url: str
    mime_type: NotRequired[str]


class AudioPart(TypedDict):
    type: Literal["audio"]
    data: bytes
    mime_type: str
    sample_rate: NotRequired[int]
    channels: NotRequired[int]


class JsonPart(TypedDict):
    type: Literal["json"]
    data: Dict[str, Any]


ContentPart = Union[TextPart, ImageBytesPart, ImageURLPart, AudioPart, JsonPart]


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    TOOL = "tool"


class Message(TypedDict):
    role: Role
    content: List[ContentPart]


# =============================
# Normalized response contracts
# =============================


class NormalizedStatus(str, Enum):
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    INCOMPLETE = "incomplete"
    FAILED = "failed"


class ToolSchema(TypedDict, total=False):
    name: str
    description: NotRequired[str]
    parameters_schema: Dict[str, Any]


class ResponseFormat(TypedDict, total=False):
    # JSON Schema to enforce structured output, if supported by provider
    json_schema: Dict[str, Any]
    # If true, providers should enforce strict adherence where possible
    strict: NotRequired[bool]


class NormalizedUsage(TypedDict, total=False):
    input_tokens: int
    output_tokens: int
    total_tokens: int
    raw_usage: NotRequired[Dict[str, Any]]


class NormalizedTextItem(TypedDict, total=False):
    type: Literal["text"]
    text: str
    index: NotRequired[int]


class NormalizedAudioItem(TypedDict, total=False):
    type: Literal["audio"]
    data: bytes
    mime_type: NotRequired[str]
    sample_rate: NotRequired[int]
    channels: NotRequired[int]
    index: NotRequired[int]


class NormalizedImageItem(TypedDict, total=False):
    type: Literal["image"]
    data: NotRequired[bytes]
    url: NotRequired[str]
    mime_type: NotRequired[str]
    index: NotRequired[int]


class NormalizedToolCallItem(TypedDict, total=False):
    type: Literal["tool_call"]
    name: str
    arguments_json: Dict[str, Any]
    id: NotRequired[str]  # Provider-specific tool call ID (e.g., for OpenAI, Anthropic)
    thought_signature: NotRequired[
        str
    ]  # Gemini-specific thought signature for multi-turn function calls


class NormalizedToolResultItem(TypedDict, total=False):
    type: Literal["tool_result"]
    name: str
    result_json: Dict[str, Any]
    is_error: NotRequired[bool]


NormalizedOutputItem = Union[
    NormalizedTextItem,
    NormalizedAudioItem,
    NormalizedImageItem,
    NormalizedToolCallItem,
    NormalizedToolResultItem,
]


class NormalizedResponse(TypedDict, total=False):
    id: str
    model: str
    status: NormalizedStatus
    output: List[NormalizedOutputItem]
    output_text: NotRequired[str]
    usage: NotRequired[NormalizedUsage]
    metadata: NotRequired[Dict[str, Any]]
    warnings: NotRequired[List[str]]
    incomplete_details: NotRequired[Dict[str, Any]]
    raw: NotRequired[Any]


__all__ = [
    "TextPart",
    "ImageBytesPart",
    "ImageURLPart",
    "AudioPart",
    "JsonPart",
    "ContentPart",
    "Role",
    "Message",
    "NormalizedStatus",
    "ToolSchema",
    "ResponseFormat",
    "NormalizedUsage",
    "NormalizedTextItem",
    "NormalizedAudioItem",
    "NormalizedImageItem",
    "NormalizedToolCallItem",
    "NormalizedToolResultItem",
    "NormalizedOutputItem",
    "NormalizedResponse",
]



import asyncio
import logging
import weakref
from typing import TYPE_CHECKING, Awaitable, Callable, Optional

from vision_agents.core.utils.utils import await_or_run, cancel_and_wait
from vision_agents.core.warmup import Warmable, WarmupCache

if TYPE_CHECKING:
    from .agents import Agent

logger = logging.getLogger(__name__)


class AgentLauncher:
    """
    Agent launcher that handles warmup and lifecycle management.

    The launcher ensures all components (LLM, TTS, STT, turn detection)
    are warmed up before the agent is launched.
    """

    def __init__(
        self,
        create_agent: Callable[..., "Agent" | Awaitable["Agent"]],
        join_call: Callable[..., None | Awaitable[None]] | None = None,
        agent_idle_timeout: float = 60.0,
        agent_idle_cleanup_interval: float = 5.0,
    ):
        """
        Initialize the agent launcher.

        Args:
            create_agent: A function that creates and returns an Agent instance
            join_call: Optional function that handles joining a call with the agent
            agent_idle_timeout: Optional timeout in seconds for agent to stay alone on the call. Default - `60.0`.
                `0` means idle agents won't leave the call until it's ended.

        """
        self.create_agent = create_agent
        self.join_call = join_call
        self._warmup_lock = asyncio.Lock()
        self._warmup_cache = WarmupCache()

        if agent_idle_timeout < 0:
            raise ValueError("agent_idle_timeout must be >= 0")
        self._agent_idle_timeout = agent_idle_timeout

        if agent_idle_cleanup_interval <= 0:
            raise ValueError("agent_idle_cleanup_interval must be > 0")
        self._agent_idle_cleanup_interval = agent_idle_cleanup_interval

        self._active_agents: weakref.WeakSet[Agent] = weakref.WeakSet()

        self._running = False
        self._cleanup_task: Optional[asyncio.Task] = None
        self._warmed_up: bool = False

    async def start(self):
        if self._running:
            raise RuntimeError("AgentLauncher is already running")
        logger.debug("Starting AgentLauncher")
        self._running = True
        await self.warmup()
        self._cleanup_task = asyncio.create_task(self._cleanup_idle_agents())
        logger.debug("AgentLauncher started")

    async def stop(self):
        logger.debug("Stopping AgentLauncher")
        self._running = False
        if self._cleanup_task:
            await cancel_and_wait(self._cleanup_task)
        logger.debug("AgentLauncher stopped")

    async def warmup(self) -> None:
        """
        Warm up all agent components.

        This method creates the agent and calls warmup() on LLM, TTS, STT,
        and turn detection components if they exist.
        """
        if self._warmed_up or self._warmup_lock.locked():
            return

        async with self._warmup_lock:
            logger.info("Creating agent...")

            # Create a dry-run Agent instance and warmup its components for the first time.
            agent: "Agent" = await await_or_run(self.create_agent)
            logger.info("Warming up agent components...")
            await self._warmup_agent(agent)
            self._warmed_up = True

            logger.info("Agent warmup completed")

    async def launch(self, **kwargs) -> "Agent":
        """
        Launch the agent.

        Args:
            **kwargs: Additional keyword arguments to pass to create_agent

        Returns:
            The Agent instance
        """
        agent: "Agent" = await await_or_run(self.create_agent, **kwargs)
        await self._warmup_agent(agent)
        self._active_agents.add(agent)
        return agent

    async def _warmup_agent(self, agent: "Agent") -> None:
        """
        Go over the Agent's dependencies and trigger `.warmup()` on them.

        It is safe to call `._warmup_agent()` multiple times.

        Args:
            agent: Agent to be warmed up

        Returns:

        """
        # Warmup tasks to run in parallel
        warmup_tasks = []

        # Warmup LLM (including Realtime)
        if agent.llm and isinstance(agent.llm, Warmable):
            warmup_tasks.append(agent.llm.warmup(self._warmup_cache))

        # Warmup TTS
        if agent.tts and isinstance(agent.tts, Warmable):
            warmup_tasks.append(agent.tts.warmup(self._warmup_cache))

        # Warmup STT
        if agent.stt and isinstance(agent.stt, Warmable):
            warmup_tasks.append(agent.stt.warmup(self._warmup_cache))

        # Warmup turn detection
        if agent.turn_detection and isinstance(agent.turn_detection, Warmable):
            warmup_tasks.append(agent.turn_detection.warmup(self._warmup_cache))

        # Warmup processors
        for processor in agent.processors:
            if isinstance(processor, Warmable):
                warmup_tasks.append(processor.warmup(self._warmup_cache))

        if warmup_tasks:
            await asyncio.gather(*warmup_tasks)

    async def _cleanup_idle_agents(self) -> None:
        if not self._agent_idle_timeout:
            return

        while self._running:
            # Collect idle agents first to close them all at once
            idle_agents = []
            for agent in self._active_agents:
                agent_idle_for = agent.idle_for()
                if agent_idle_for >= self._agent_idle_timeout:
                    logger.info(
                        f'Agent with user_id "{agent.agent_user.id}" is idle for {round(agent_idle_for, 2)}s, '
                        f"closing it after {self._agent_idle_timeout}s timeout"
                    )
                    idle_agents.append(agent)

            if idle_agents:
                coros = [asyncio.shield(a.close()) for a in idle_agents]
                result = await asyncio.shield(
                    asyncio.gather(*coros, return_exceptions=True)
                )
                for agent, r in zip(idle_agents, result):
                    if isinstance(r, Exception):
                        logger.error(
                            f"Failed to close idle agent with user_id {agent.agent_user.id}",
                            exc_info=r,
                        )

            await asyncio.sleep(self._agent_idle_cleanup_interval)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()



"""Video frame utilities."""

import io
import logging

import av
from PIL import Image
from PIL.Image import Resampling

logger = logging.getLogger(__name__)


def ensure_even_dimensions(frame: av.VideoFrame) -> av.VideoFrame:
    """
    Ensure frame has even dimensions for H.264 yuv420p encoding.

    Crops 1 pixel from right/bottom edge if width/height is odd.
    """
    needs_width_adjust = frame.width % 2 != 0
    needs_height_adjust = frame.height % 2 != 0

    if not needs_width_adjust and not needs_height_adjust:
        return frame

    new_width = frame.width - (1 if needs_width_adjust else 0)
    new_height = frame.height - (1 if needs_height_adjust else 0)

    # Convert to numpy, crop (slice), convert back - faster than reformat which rescales
    arr = frame.to_ndarray(format="rgb24")
    cropped_arr = arr[:new_height, :new_width]
    cropped = av.VideoFrame.from_ndarray(cropped_arr, format="rgb24")
    cropped.pts = frame.pts
    if frame.time_base is not None:
        cropped.time_base = frame.time_base

    return cropped


def frame_to_jpeg_bytes(
    frame: av.VideoFrame, target_width: int, target_height: int, quality: int = 85
) -> bytes:
    """
    Convert a video frame to JPEG bytes with resizing.

    Args:
        frame: an instance of `av.VideoFrame`.
        target_width: target width in pixels.
        target_height: target height in pixels.
        quality: JPEG quality. Default is 85.

    Returns: frame as JPEG bytes.

    """
    # Convert frame to a PIL image
    img = frame.to_image()

    # Calculate scaling to maintain aspect ratio
    src_width, src_height = img.size
    # Calculate scale factor (fit within target dimensions)
    scale = min(target_width / src_width, target_height / src_height)
    new_width = int(src_width * scale)
    new_height = int(src_height * scale)

    # Resize with aspect ratio maintained
    resized = img.resize((new_width, new_height), Resampling.LANCZOS)

    # Save as JPEG with quality control
    buf = io.BytesIO()
    resized.save(buf, "JPEG", quality=quality, optimize=True)
    return buf.getvalue()


def frame_to_png_bytes(frame: av.VideoFrame) -> bytes:
    """
    Convert a video frame to PNG bytes.

    Args:
        frame: Video frame object that can be converted to an image

    Returns:
        PNG bytes of the frame, or empty bytes if conversion fails
    """
    if hasattr(frame, "to_image"):
        img = frame.to_image()
    else:
        arr = frame.to_ndarray(format="rgb24")
        img = Image.fromarray(arr)

    buf = io.BytesIO()
    img.save(buf, format="PNG")
    return buf.getvalue()


def resize_frame(self, frame: av.VideoFrame) -> av.VideoFrame:
    """
    Resizes a video frame to target dimensions while maintaining the aspect ratio. The method centers the resized
    image on a black background if the target dimensions do not match the original aspect ratio.

    Parameters:
        frame (av.VideoFrame): The input video frame to be resized.

    Returns:
        av.VideoFrame: The output video frame after resizing, maintaining the original aspect ratio.

    Raises:
        None
    """
    img = frame.to_image()

    # Calculate scaling to maintain aspect ratio
    src_width, src_height = img.size
    target_width, target_height = self.width, self.height

    # Calculate scale factor (fit within target dimensions)
    scale = min(target_width / src_width, target_height / src_height)
    new_width = int(src_width * scale)
    new_height = int(src_height * scale)

    # Resize with aspect ratio maintained
    resized = img.resize((new_width, new_height), Image.Resampling.LANCZOS)

    # Create black background at target resolution
    result = Image.new("RGB", (target_width, target_height), (0, 0, 0))

    # Paste resized image centered
    x_offset = (target_width - new_width) // 2
    y_offset = (target_height - new_height) // 2
    result.paste(resized, (x_offset, y_offset))

    return av.VideoFrame.from_image(result)



from .llm import LLM, AudioLLM, VideoLLM, OmniLLM
from .realtime import Realtime
from .function_registry import FunctionRegistry, function_registry

__all__ = [
    "LLM",
    "AudioLLM",
    "VideoLLM",
    "OmniLLM",
    "Realtime",
    "FunctionRegistry",
    "function_registry",
]



"""Stream Edge Transport Package.

This package provides edge transport abstraction for Stream Agents.
"""

from vision_agents.core.edge.edge_transport import EdgeTransport
from vision_agents.core.edge import sfu_events

__all__ = ["EdgeTransport", "sfu_events"]



"""Remote MCP server connection using HTTP Streamable transport."""

from datetime import timedelta
from typing import Optional, Dict, Callable
from urllib.parse import urlparse

from mcp import ClientSession
from mcp.client.streamable_http import streamablehttp_client

from .mcp_base import MCPBaseServer


class MCPServerRemote(MCPBaseServer):
    """Remote MCP server connection using HTTP Streamable transport."""

    def __init__(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        timeout: float = 30.0,
        session_timeout: float = 300.0,
    ):
        """Initialize the remote MCP server connection.

        Args:
            url: URL of the MCP server (e.g., "http://localhost:8001/mcp")
            headers: Optional HTTP headers to include in requests
            timeout: Connection timeout in seconds
            session_timeout: How long an established MCP session can sit idle with no tool calls, no traffic (in seconds)
        """
        super().__init__(session_timeout)
        self.url = url
        self.headers = headers or {}
        self.timeout = timeout
        self._client_context: Optional[object] = None  # AsyncGeneratorContextManager
        self._session_context: Optional[object] = None  # ClientSession context manager
        self._get_session_id_cb: Optional[Callable[[], Optional[str]]] = None

        # Validate URL
        parsed = urlparse(url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(f"Invalid URL: {url}")

    async def connect(self) -> None:
        """Connect to the remote MCP server."""
        if self._is_connected:
            self.logger.warning("Already connected to MCP server")
            return

        try:
            self.logger.info(f"Connecting to remote MCP server at {self.url}")

            # Create the HTTP client context
            self._client_context = streamablehttp_client(  # type: ignore[assignment]
                self.url, headers=self.headers, timeout=timedelta(seconds=self.timeout)
            )

            # Enter the context to get the read/write streams and session ID callback
            (
                read,
                write,
                self._get_session_id_cb,
            ) = await self._client_context.__aenter__()  # type: ignore[attr-defined]

            # Create the client session context manager
            self._session_context = ClientSession(read, write)  # type: ignore[assignment]

            # Enter the session context and get the actual session
            self._session = await self._session_context.__aenter__()  # type: ignore[attr-defined]

            # Initialize the connection
            await self._session.initialize()

            self._is_connected = True
            await self._update_activity()
            await self._start_timeout_monitor()

            # Log session ID if available
            if self._get_session_id_cb is not None:
                try:
                    session_id = self._get_session_id_cb()
                    self.logger.info(
                        f"Successfully connected to remote MCP server at {self.url} (session: {session_id})"
                    )
                except Exception as e:
                    self.logger.info(
                        f"Successfully connected to remote MCP server at {self.url} (session ID unavailable: {e})"
                    )
            else:
                self.logger.info(
                    f"Successfully connected to remote MCP server at {self.url}"
                )

        except Exception as e:
            self.logger.error(f"Failed to connect to remote MCP server: {e}")
            # Clean up any partial connection state
            await self._cleanup_connection()
            raise

    async def disconnect(self) -> None:
        """Disconnect from the remote MCP server."""
        if not self._is_connected:
            return

        try:
            self.logger.info("Disconnecting from remote MCP server")

            # Stop timeout monitoring
            await self._stop_timeout_monitor()

            # Clean up the connection
            await self._cleanup_connection()

            self._is_connected = False
            self.logger.info("Disconnected from remote MCP server")

        except Exception as e:
            self.logger.error(f"Error disconnecting from remote MCP server: {e}")
            self._is_connected = False

    async def _cleanup_connection(self) -> None:
        """Clean up the MCP connection resources."""
        # Close the session context
        if self._session_context:
            try:
                await self._session_context.__aexit__(None, None, None)  # type: ignore[attr-defined]
            except Exception as e:
                self.logger.warning(f"Error closing MCP session context: {e}")
            self._session_context = None

        # Close the client context
        if self._client_context:
            try:
                await self._client_context.__aexit__(None, None, None)  # type: ignore[attr-defined]
            except Exception as e:
                self.logger.warning(f"Error closing MCP client context: {e}")
            self._client_context = None

        self._session = None
        self._get_session_id_cb = None

    async def __aenter__(self):
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.disconnect()

    def __repr__(self) -> str:
        """String representation of the remote MCP server."""
        return f"MCPServerRemote(url='{self.url}', connected={self._is_connected})"



from __future__ import annotations

from typing import (
    Optional,
)

from getstream.video.rtc.track_util import PcmData
from vision_agents.core.edge.types import Participant


import abc
import logging
import uuid

from . import events, OmniLLM


logger = logging.getLogger(__name__)


class Realtime(OmniLLM):
    """
    Realtime is an abstract base class for LLMs that can receive audio and video

    Example:

        llm = Realtime()
        llm.connect()
        llm.simple_response("what do you see?")

    Emits the following events:

    TODO: document/ evaluate how many events we want/ need...
        - Transcript incoming audio
        - Transcript outgoing audio

    """

    fps: int = 1
    session_id: str  # UUID to identify this session

    def __init__(
        self,
        fps: int = 1,  # the number of video frames per second to send (for implementations that support setting fps)
    ):
        super().__init__()
        self.connected = False

        self.provider_name = "realtime_base"
        self.session_id = str(uuid.uuid4())
        self.fps = fps
        # Store current participant for user speech transcription events
        self._current_participant: Optional[Participant] = None

    @abc.abstractmethod
    async def connect(self): ...

    @abc.abstractmethod
    async def simple_audio_response(
        self, pcm: PcmData, participant: Optional[Participant] = None
    ): ...

    async def stop_watching_video_track(self) -> None:
        """Optionally overridden by providers that support video input."""
        return None

    def _emit_connected_event(self, session_config=None, capabilities=None):
        """Emit a structured connected event."""
        self.connected = True
        # Mark ready when connected if provider uses base emitter
        try:
            self._ready_event.set()  # type: ignore[attr-defined]
        except Exception:
            pass
        event = events.RealtimeConnectedEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            session_config=session_config,
            capabilities=capabilities,
        )
        self.events.send(event)

    def _emit_disconnected_event(self, reason=None, was_clean=True):
        """Emit a structured disconnected event."""
        self.connected = False
        event = events.RealtimeDisconnectedEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            reason=reason,
            was_clean=was_clean,
        )
        self.events.send(event)

    def _emit_audio_input_event(self, audio_data: PcmData, user_metadata=None):
        """Emit a structured audio input event."""
        event = events.RealtimeAudioInputEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            data=audio_data,
            participant=user_metadata,
        )
        self.events.send(event)

    # TODO: discussion around event vs output_track... why do we have both?
    def _emit_audio_output_event(
        self, audio_data: PcmData, response_id=None, user_metadata=None
    ):
        """Emit a structured audio output event."""
        event = events.RealtimeAudioOutputEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            data=audio_data,
            response_id=response_id,
            participant=user_metadata,
        )
        self.events.send(event)

    def _emit_response_event(
        self,
        text,
        response_id=None,
        is_complete=True,
        conversation_item_id=None,
        user_metadata=None,
    ):
        """Emit a structured response event."""
        event = events.RealtimeResponseEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            text=text,
            response_id=response_id,
            is_complete=is_complete,
            conversation_item_id=conversation_item_id,
            participant=user_metadata,
        )
        self.events.send(event)

    def _emit_conversation_item_event(
        self, item_id, item_type, status, role, content=None, user_metadata=None
    ):
        """Emit a structured conversation item event."""
        event = events.RealtimeConversationItemEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            item_id=item_id,
            item_type=item_type,
            status=status,
            role=role,
            content=content,
            participant=user_metadata,
        )
        self.events.send(event)

    def _emit_error_event(self, error, context="", user_metadata=None):
        """Emit a structured error event."""
        event = events.RealtimeErrorEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            error=error,
            context=context,
            participant=user_metadata,
        )
        self.events.send(event)

    @abc.abstractmethod
    async def close(self):
        raise NotImplementedError("llm.close isn't implemented")

    def _emit_user_speech_transcription(self, text: str, original=None):
        """Emit a user speech transcription event with participant info."""
        event = events.RealtimeUserSpeechTranscriptionEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            text=text,
            original=original,
            participant=self._current_participant,
        )
        self.events.send(event)

    def _emit_agent_speech_transcription(self, text: str, original=None):
        """Emit an agent speech transcription event."""
        event = events.RealtimeAgentSpeechTranscriptionEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            text=text,
            original=original,
        )
        self.events.send(event)



from dataclasses import dataclass
from typing import (
    Any,
    Optional,
    Protocol,
    runtime_checkable,
)


from getstream.video.rtc import PcmData
from pyee.asyncio import AsyncIOEventEmitter


@dataclass
class User:
    id: Optional[str] = ""
    name: Optional[str] = ""
    image: Optional[str] = ""


@dataclass
class Participant:
    original: Any
    user_id: str


class Connection(AsyncIOEventEmitter):
    """
    To standardize we need to have a method to close
    and a way to receive a callback when the call is ended
    In the future we might want to forward more events
    """

    async def close(self):
        pass


@runtime_checkable
class OutputAudioTrack(Protocol):
    """
    A protocol describing an output audio track, the actual implementation depends on the edge transported used
    eg. getstream.video.rtc.audio_track.AudioStreamTrack
    """

    async def write(self, data: PcmData) -> None: ...

    def stop(self) -> None: ...

    async def flush(self) -> None: ...



from typing import Dict, Any

import httpx


async def get_weather_by_location(location: str) -> Dict[str, Any]:
    """
    Get current weather for a location using Open-Meteo API.

    Args:
        location: Name of the location (city, place, etc.)

    Returns:
        Weather data dictionary containing current weather information

    Raises:
        ValueError: If location not found or API response is invalid
        httpx.HTTPError: If API request fails
    """
    async with httpx.AsyncClient(timeout=10.0) as client:
        # Get geocoding data for the location
        geo_response = await client.get(
            "https://geocoding-api.open-meteo.com/v1/search", params={"name": location}
        )
        geo_response.raise_for_status()
        geo_data = geo_response.json()

        if not geo_data.get("results"):
            raise ValueError(f"Location '{location}' not found")

        # Get weather for the location
        lat = geo_data["results"][0]["latitude"]
        lon = geo_data["results"][0]["longitude"]

        weather_response = await client.get(
            "https://api.open-meteo.com/v1/forecast",
            params={"latitude": lat, "longitude": lon, "current_weather": True},
        )
        weather_response.raise_for_status()
        weather_data = weather_response.json()

        return weather_data



from vision_agents.core.edge.types import User

from vision_agents.core.agents import Agent

from vision_agents.core.cli.cli_runner import cli
from vision_agents.core.agents.agent_launcher import AgentLauncher

__all__ = ["Agent", "User", "cli", "AgentLauncher"]



import abc
import logging
import uuid
from typing import Optional
from getstream.video.rtc.track_util import PcmData

from ..edge.types import Participant
from vision_agents.core.events.manager import EventManager
from . import events
from .events import TranscriptResponse
from ..turn_detection import TurnEndedEvent, TurnStartedEvent

logger = logging.getLogger(__name__)


class STT(abc.ABC):
    """
    Abstract base class for Speech-to-Text implementations.

    Subclasses implement this and have to call
    - _emit_partial_transcript_event
    - _emit_transcript_event
    - _emit_error_event for temporary errors

    process_audio is currently called every 20ms. The integration with turn keeping could be improved
    """

    closed: bool = False
    started: bool = False
    turn_detection: bool = False  # if the STT supports turn detection

    def __init__(
        self,
        provider_name: Optional[str] = None,
    ):
        self.session_id = str(uuid.uuid4())
        self.provider_name = provider_name or self.__class__.__name__

        self.events = EventManager()
        self.events.register(TurnEndedEvent)
        self.events.register(TurnStartedEvent)
        self.events.register_events_from_module(events, ignore_not_compatible=True)

    def _emit_transcript_event(
        self,
        text: str,
        participant: Participant,
        response: TranscriptResponse,
    ):
        """
        Emit a final transcript event with structured data.

        Args:
            text: The transcribed text.
            participant: Participant metadata.
            response: Transcription response metadata.
        """
        self.events.send(
            events.STTTranscriptEvent(
                session_id=self.session_id,
                plugin_name=self.provider_name,
                text=text,
                participant=participant,
                response=response,
            )
        )

    def _emit_turn_ended_event(
        self,
        participant: Participant,
        eager_end_of_turn: bool = False,
        confidence: Optional[float] = None,
    ):
        if confidence is None:
            confidence = 0.5
        self.events.send(
            TurnEndedEvent(
                session_id=self.session_id,
                plugin_name=self.provider_name,
                participant=participant,
                eager_end_of_turn=eager_end_of_turn,
                confidence=confidence,
            )
        )

    def _emit_turn_started_event(
        self,
        participant: Participant,
        confidence: Optional[float] = None,
    ):
        if confidence is None:
            confidence = 0.5
        event = TurnStartedEvent(
            session_id=self.session_id,
            plugin_name=self.provider_name,
            participant=participant,
            confidence=confidence,
        )
        self.events.send(event)

    def _emit_partial_transcript_event(
        self,
        text: str,
        participant: Participant,
        response: TranscriptResponse,
    ):
        """
        Emit a partial transcript event with structured data.

        Args:
            text: The partial transcribed text.
            participant: Participant metadata.
            response: Transcription response metadata.
        """
        self.events.send(
            events.STTPartialTranscriptEvent(
                session_id=self.session_id,
                plugin_name=self.provider_name,
                text=text,
                participant=participant,
                response=response,
            )
        )

    def _emit_error_event(
        self,
        error: Exception,
        context: str = "",
        participant: Optional[Participant] = None,
    ):
        """
        Emit an error event. Note this should only be emitted for temporary errors.
        Permanent errors due to config etc should be directly raised
        """
        self.events.send(
            events.STTErrorEvent(
                session_id=self.session_id,
                plugin_name=self.provider_name,
                error=error,
                context=context,
                participant=participant,
                error_code=getattr(error, "error_code", None),
                is_recoverable=not isinstance(error, (SystemExit, KeyboardInterrupt)),
            )
        )

    @abc.abstractmethod
    async def process_audio(
        self,
        pcm_data: PcmData,
        participant: Optional[Participant] = None,
    ):
        pass

    async def start(self):
        if self.started:
            raise ValueError("STT is already started, dont call this method twice")
        self.started = True

    async def clear(self):
        """Clear any pending audio or state. Override in subclasses if needed."""
        pass

    async def close(self):
        self.closed = True



"""MCP (Model Context Protocol) integration for Stream Agents."""

from .mcp_server_remote import MCPServerRemote
from .mcp_server_local import MCPServerLocal
from .mcp_base import MCPBaseServer
from .tool_converter import MCPToolConverter
from .mcp_manager import MCPManager

__all__ = [
    "MCPServerRemote",
    "MCPServerLocal",
    "MCPBaseServer",
    "MCPToolConverter",
    "MCPManager",
]



"""
Function registry for managing available functions that can be called by LLMs.
"""

from __future__ import annotations

import inspect
from typing import Any, Callable, Dict, List, Optional, Type, Union, get_type_hints
from dataclasses import dataclass
from enum import Enum

from .llm_types import ToolSchema


@dataclass
class FunctionParameter:
    """Represents a parameter of a function."""

    name: str
    type: Type
    description: Optional[str] = None
    required: bool = True
    default: Any = None


@dataclass
class FunctionDefinition:
    """Represents a complete function definition."""

    name: str
    description: str
    parameters: List[FunctionParameter]
    function: Callable
    returns: Optional[Type] = None


@dataclass
class _ExplicitSchemaFunction:
    """Represents a function with an explicitly provided JSON schema (e.g., MCP tools)."""

    name: str
    description: str
    parameters_schema: Dict[str, Any]
    function: Callable


class FunctionRegistry:
    """Registry for managing available functions that can be called by LLMs."""

    def __init__(self):
        self._functions: Dict[str, FunctionDefinition | _ExplicitSchemaFunction] = {}

    def register(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        parameters_schema: Optional[Dict[str, Any]] = None,
    ) -> Callable:
        """
        Decorator to register a function with the registry.

        Args:
            name: Optional custom name for the function. If not provided, uses the function name.
            description: Optional description for the function. If not provided, uses the docstring.
            parameters_schema: Optional explicit JSON schema for parameters. If provided, this is
                used instead of inferring from type hints. Useful for MCP tools or external APIs.

        Returns:
            Decorator function.
        """

        def decorator(func: Callable) -> Callable:
            func_name = name or func.__name__
            func_description = description or func.__doc__ or ""

            # If explicit schema provided, use it directly
            if parameters_schema is not None:
                self._functions[func_name] = _ExplicitSchemaFunction(
                    name=func_name,
                    description=func_description,
                    parameters_schema=parameters_schema,
                    function=func,
                )
                return func

            # Otherwise, extract from type hints
            type_hints = get_type_hints(func)
            sig = inspect.signature(func)

            parameters = []
            for param_name, param in sig.parameters.items():
                if param_name == "self":
                    continue

                param_type = type_hints.get(param_name, type(None))
                param_description = None

                # Check if there's a docstring with parameter descriptions
                if func.__doc__:
                    # Simple extraction of parameter descriptions from docstring
                    # This is a basic implementation - could be enhanced with proper parsing
                    pass

                parameters.append(
                    FunctionParameter(
                        name=param_name,
                        type=param_type,
                        description=param_description,
                        required=param.default == inspect.Parameter.empty,
                        default=param.default
                        if param.default != inspect.Parameter.empty
                        else None,
                    )
                )

            # Determine return type
            return_type = type_hints.get("return", None)

            function_def = FunctionDefinition(
                name=func_name,
                description=func_description,
                parameters=parameters,
                function=func,
                returns=return_type,
            )

            self._functions[func_name] = function_def
            return func

        return decorator

    def get_function(
        self, name: str
    ) -> Optional[FunctionDefinition | _ExplicitSchemaFunction]:
        """Get a function definition by name."""
        return self._functions.get(name)

    def list_functions(self) -> List[str]:
        """Get a list of all registered function names."""
        return list(self._functions.keys())

    def get_tool_schemas(self) -> List[ToolSchema]:
        """Get tool schemas for all registered functions."""
        schemas = []
        for func_def in self._functions.values():
            if isinstance(func_def, _ExplicitSchemaFunction):
                # Use the explicitly provided schema
                schemas.append(
                    ToolSchema(
                        name=func_def.name,
                        description=func_def.description,
                        parameters_schema=func_def.parameters_schema,
                    )
                )
            else:
                schema = self._function_to_tool_schema(func_def)
                schemas.append(schema)
        return schemas

    def call_function(self, name: str, arguments: Dict[str, Any]) -> Any:
        """
        Call a registered function with the given arguments.

        Args:
            name: Name of the function to call.
            arguments: Dictionary of arguments to pass to the function.

        Returns:
            Result of the function call.

        Raises:
            KeyError: If the function is not registered.
            TypeError: If the arguments don't match the function signature.
        """
        if name not in self._functions:
            raise KeyError(f"Function '{name}' is not registered")

        func_def = self._functions[name]

        # For explicit schema functions, just call directly (validation is external)
        if isinstance(func_def, _ExplicitSchemaFunction):
            return func_def.function(**arguments)

        # Validate required parameters for introspected functions
        for param in func_def.parameters:
            if param.required and param.name not in arguments:
                raise TypeError(
                    f"Missing required parameter '{param.name}' for function '{name}'"
                )

        # Call the function with the provided arguments
        return func_def.function(**arguments)

    def get_callable(self, name: str) -> Callable:
        """
        Get the callable function by name.

        Args:
            name: Name of the function

        Returns:
            The callable function

        Raises:
            KeyError: If the function is not registered
        """
        if name not in self._functions:
            raise KeyError(f"Function '{name}' is not registered")

        return self._functions[name].function

    def _function_to_tool_schema(self, func_def: FunctionDefinition) -> ToolSchema:
        """Convert a function definition to a tool schema."""
        properties = {}
        required = []

        for param in func_def.parameters:
            param_schema = self._type_to_json_schema(param.type)
            if param.description:
                param_schema["description"] = param.description

            properties[param.name] = param_schema

            if param.required:
                required.append(param.name)

        schema = {"type": "object", "properties": properties}

        if required:
            schema["required"] = required

        return ToolSchema(
            name=func_def.name,
            description=func_def.description,
            parameters_schema=schema,
        )

    def _type_to_json_schema(self, type_hint: Type) -> Dict[str, Any]:
        """Convert a Python type hint to a JSON schema."""
        # Handle basic types
        if type_hint is str:
            return {"type": "string"}
        elif type_hint is int:
            return {"type": "integer"}
        elif type_hint is float:
            return {"type": "number"}
        elif type_hint is bool:
            return {"type": "boolean"}
        elif type_hint is list or type_hint is List:
            return {"type": "array"}
        elif type_hint is dict or type_hint is Dict:
            return {"type": "object"}

        # Handle Optional types
        if hasattr(type_hint, "__origin__") and type_hint.__origin__ is Union:
            # Check if it's Optional (Union[SomeType, None])
            args = type_hint.__args__
            if len(args) == 2 and type(None) in args:
                non_none_type = args[0] if args[1] is type(None) else args[1]
                return self._type_to_json_schema(non_none_type)

        # Handle List types
        if hasattr(type_hint, "__origin__") and type_hint.__origin__ is list:
            if hasattr(type_hint, "__args__") and type_hint.__args__:
                item_type = type_hint.__args__[0]
                return {"type": "array", "items": self._type_to_json_schema(item_type)}
            return {"type": "array"}

        # Handle Dict types
        if hasattr(type_hint, "__origin__") and type_hint.__origin__ is dict:
            return {"type": "object"}

        # Handle Enum types
        if inspect.isclass(type_hint) and issubclass(type_hint, Enum):
            return {"type": "string", "enum": [e.value for e in type_hint]}

        # Default fallback
        return {"type": "string"}


# Global registry instance
function_registry = FunctionRegistry()



"""Utility to generate static SFU event dataclass definitions.

This script inspects ``events_pb2`` at runtime and materialises a static Python
module containing dataclass wrappers for every protobuf message emitted by the
SFU. Run this script whenever the upstream protobuf schema changes.
"""

from __future__ import annotations

import pathlib
from typing import (
    Dict as TypingDict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
)

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message

from getstream.video.rtc.pb.stream.video.sfu.event import events_pb2
from getstream.video.rtc.pb.stream.video.sfu.models import models_pb2


HEADER_LINES: Sequence[str] = (
    "from __future__ import annotations",
    "",
    "import uuid",
    "from dataclasses import dataclass, field",
    "from datetime import datetime, timezone",
    "from typing import Any, Dict, List, Optional",
    "",
    "from dataclasses_json import DataClassJsonMixin",
    "from google.protobuf.json_format import MessageToDict",
    "from getstream.video.rtc.pb.stream.video.sfu.event import events_pb2",
    "from vision_agents.core.events.base import BaseEvent",
    "",
    "# Note: For enum fields typed as 'int', use the corresponding enum from:",
    "# getstream.video.rtc.pb.stream.video.sfu.models.models_pb2",
    "# Available enums: TrackType, ConnectionQuality, ErrorCode, PeerType, etc.",
    "",
    "",
    "def _to_dict(message) -> Dict[str, Any]:",
    "    try:",
    "        return MessageToDict(message, preserving_proto_field_name=True)",
    "    except Exception:",
    "        return {}",
    "",
)


def _iter_protobuf_messages() -> Iterable[Tuple[str, Type[Message]]]:
    for name in sorted(dir(events_pb2)):
        attr = getattr(events_pb2, name)
        if isinstance(attr, type) and issubclass(attr, Message):
            yield name, attr


def _class_name(proto_name: str) -> str:
    return proto_name if proto_name.endswith("Event") else proto_name + "Event"


def _collect_message_types() -> TypingDict[str, Type[Message]]:
    """Collect all message types referenced in event fields (recursively)."""
    message_types: TypingDict[str, Type[Message]] = {}
    to_process: Set[str] = set()

    # Collect all message types from events
    for proto_name, message_cls in _iter_protobuf_messages():
        for field_desc in message_cls.DESCRIPTOR.fields:
            if field_desc.type == FieldDescriptor.TYPE_MESSAGE:
                message_type_name = field_desc.message_type.full_name
                if message_type_name.startswith("stream.video.sfu.model"):
                    class_name = message_type_name.split(".")[-1]
                    to_process.add(class_name)

    # Process messages and their nested message types recursively
    processed: Set[str] = set()
    while to_process:
        class_name = to_process.pop()
        if class_name in processed:
            continue
        processed.add(class_name)

        if hasattr(models_pb2, class_name):
            message_cls = getattr(models_pb2, class_name)
            message_types[class_name] = message_cls

            # Check for nested message types
            for field_desc in message_cls.DESCRIPTOR.fields:
                if field_desc.type == FieldDescriptor.TYPE_MESSAGE:
                    nested_type_name = field_desc.message_type.full_name
                    if nested_type_name.startswith("stream.video.sfu.model"):
                        nested_class_name = nested_type_name.split(".")[-1]
                        if nested_class_name not in processed:
                            to_process.add(nested_class_name)

    return message_types


def _get_message_type_name(field_descriptor: FieldDescriptor) -> Optional[str]:
    """Get the wrapper class name for a message type field."""
    if field_descriptor.type == FieldDescriptor.TYPE_MESSAGE:
        message_type_name = field_descriptor.message_type.full_name

        # Map known message types to their wrapper classes
        if message_type_name.startswith("stream.video.sfu.model"):
            class_name = message_type_name.split(".")[-1]
            # Return the class name which will be defined at the top
            return class_name

    return None


def _get_enum_type_name(field_descriptor: FieldDescriptor) -> Optional[str]:
    """Get the enum class name for an enum type field for documentation purposes."""
    if field_descriptor.type == FieldDescriptor.TYPE_ENUM:
        enum_type_name = field_descriptor.enum_type.full_name
        # Extract just the class name from the full name (e.g., stream.video.sfu.models.TrackType -> TrackType)
        if enum_type_name.startswith("stream.video.sfu.models."):
            return enum_type_name.split(".")[-1]
    return None


def _get_python_type_from_protobuf_field(field_descriptor: FieldDescriptor) -> str:
    """Determine Python type from protobuf field descriptor.

    Maps protobuf field types to their corresponding Python types.
    All fields are returned as Optional since we want optional semantics.
    """
    # Map protobuf types to Python types
    type_map = {
        FieldDescriptor.TYPE_DOUBLE: "float",
        FieldDescriptor.TYPE_FLOAT: "float",
        FieldDescriptor.TYPE_INT64: "int",
        FieldDescriptor.TYPE_UINT64: "int",
        FieldDescriptor.TYPE_INT32: "int",
        FieldDescriptor.TYPE_FIXED64: "int",
        FieldDescriptor.TYPE_FIXED32: "int",
        FieldDescriptor.TYPE_BOOL: "bool",
        FieldDescriptor.TYPE_STRING: "str",
        FieldDescriptor.TYPE_BYTES: "bytes",
        FieldDescriptor.TYPE_UINT32: "int",
        FieldDescriptor.TYPE_SFIXED32: "int",
        FieldDescriptor.TYPE_SFIXED64: "int",
        FieldDescriptor.TYPE_SINT32: "int",
        FieldDescriptor.TYPE_SINT64: "int",
    }

    # Handle repeated fields (lists)
    if field_descriptor.is_repeated:
        # For enum types in repeated fields - use int, documented in docstring
        if field_descriptor.type == FieldDescriptor.TYPE_ENUM:
            return "Optional[List[int]]"
        # For message types in repeated fields
        elif field_descriptor.type == FieldDescriptor.TYPE_MESSAGE:
            message_type = _get_message_type_name(field_descriptor)
            base_type = message_type if message_type else "Any"
            return f"Optional[List[{base_type}]]"
        # For scalar types in repeated fields
        else:
            base_type = type_map.get(field_descriptor.type, "Any")
            return f"Optional[List[{base_type}]]"

    # Handle message types (nested protobuf messages)
    if field_descriptor.type == FieldDescriptor.TYPE_MESSAGE:
        message_type = _get_message_type_name(field_descriptor)
        return f"Optional[{message_type}]" if message_type else "Optional[Any]"

    # Handle enum types - use int, documented in docstring
    if field_descriptor.type == FieldDescriptor.TYPE_ENUM:
        return "Optional[int]"

    # Handle scalar types - all made optional
    base_type = type_map.get(field_descriptor.type, "Any")
    return f"Optional[{base_type}]"


def _render_message_wrapper(class_name: str, message_cls: Type[Message]) -> List[str]:
    """Generate a dataclass wrapper for a protobuf message type (like Participant)."""
    lines = ["@dataclass", f"class {class_name}(DataClassJsonMixin):"]

    # Build docstring with enum field documentation
    docstring_lines = [f"Wrapper for {message_cls.DESCRIPTOR.full_name}."]
    field_descriptors = message_cls.DESCRIPTOR.fields
    enum_fields = []
    for field_desc in field_descriptors:
        if field_desc.type == FieldDescriptor.TYPE_ENUM:
            enum_type_name = _get_enum_type_name(field_desc)
            if enum_type_name:
                enum_fields.append((field_desc.name, enum_type_name))

    if enum_fields:
        docstring_lines.append("")
        docstring_lines.append("Enum fields (use values from models_pb2):")
        for field_name, enum_name in enum_fields:
            docstring_lines.append(f"    - {field_name}: {enum_name}")

    # Add docstring
    if len(docstring_lines) == 1:
        lines.append(f'    """{docstring_lines[0]}"""')
    else:
        lines.append('    """' + docstring_lines[0])
        for line in docstring_lines[1:]:
            lines.append("    " + line)
        lines.append('    """')

    if not field_descriptors:
        lines.append("    pass")
        lines.append("")
        return lines

    # Generate fields with proper types
    for field_desc in field_descriptors:
        field_name = field_desc.name
        python_type = _get_python_type_from_protobuf_field(field_desc)

        # Use proper default for optional fields
        lines.append(f"    {field_name}: {python_type} = None")

    lines.append("")
    lines.append("    @classmethod")
    lines.append(f"    def from_proto(cls, proto_obj) -> '{class_name}':")
    lines.append(f'        """Create from protobuf {class_name}."""')
    lines.append("        if proto_obj is None:")
    lines.append("            return cls()")
    lines.append("        return cls(")

    # Generate field assignments
    for i, field_desc in enumerate(field_descriptors):
        field_name = field_desc.name
        comma = "," if i < len(field_descriptors) - 1 else ""

        # Handle different field types
        if field_desc.type == FieldDescriptor.TYPE_MESSAGE and field_desc.is_repeated:
            # Repeated message fields
            message_type = _get_message_type_name(field_desc)
            if message_type:
                lines.append(
                    f"            {field_name}=[{message_type}.from_proto(item) for item in proto_obj.{field_name}]{comma}"
                )
            else:
                lines.append(
                    f"            {field_name}=list(proto_obj.{field_name}){comma}"
                )
        elif field_desc.type == FieldDescriptor.TYPE_MESSAGE:
            # Single message field
            message_type = _get_message_type_name(field_desc)
            if message_type:
                lines.append(
                    f"            {field_name}={message_type}.from_proto(proto_obj.{field_name}) if proto_obj.HasField('{field_name}') else None{comma}"
                )
            else:
                lines.append(
                    f"            {field_name}=proto_obj.{field_name} if proto_obj.HasField('{field_name}') else None{comma}"
                )
        elif field_desc.is_repeated:
            # Repeated scalar/enum fields - convert to list of ints for enums
            lines.append(
                f"            {field_name}=list(proto_obj.{field_name}){comma}"
            )
        else:
            # Regular scalar/enum fields
            lines.append(f"            {field_name}=proto_obj.{field_name}{comma}")

    lines.append("        )")
    lines.append("")
    return lines


def _render_class(proto_name: str, message_cls: Type[Message]) -> List[str]:
    class_name = _class_name(proto_name)
    event_type = message_cls.DESCRIPTOR.full_name

    # Get field descriptors for this message
    field_descriptors = message_cls.DESCRIPTOR.fields

    lines = ["@dataclass", f"class {class_name}(BaseEvent):"]
    lines.append(
        f'    """Dataclass event for {message_cls.__module__}.{message_cls.__name__}."""'
    )

    # Override type field with the specific event type
    lines.append(f'    type: str = field(default="{event_type}", init=False)')

    # Add payload field (optional to match BaseEvent pattern)
    lines.append(
        f"    payload: Optional[events_pb2.{proto_name}] = field(default=None, repr=False)"
    )

    # Add property fields for each protobuf field (skip fields that conflict with BaseEvent)
    base_event_fields = {"type", "event_id", "timestamp", "session_id", "user_metadata"}
    for field_desc in field_descriptors:
        field_name = field_desc.name
        if (
            field_name in base_event_fields
        ):  # Skip fields that conflict with BaseEvent fields
            continue
        type_hint = _get_python_type_from_protobuf_field(field_desc)
        lines.append("")
        lines.append("    @property")
        lines.append(f"    def {field_name}(self) -> {type_hint}:")

        # Build docstring with enum information if applicable
        docstring = f"Access {field_name} field from the protobuf payload."
        if field_desc.type == FieldDescriptor.TYPE_ENUM:
            enum_type_name = _get_enum_type_name(field_desc)
            if enum_type_name:
                docstring += f" Use models_pb2.{enum_type_name} enum."

        lines.append(f'        """{docstring}"""')
        lines.append("        if self.payload is None:")
        lines.append("            return None")

        # Handle message type fields - wrap them in our dataclass
        if field_desc.type == FieldDescriptor.TYPE_MESSAGE:
            message_type = _get_message_type_name(field_desc)
            if message_type:
                if field_desc.is_repeated:
                    lines.append(
                        f"        proto_list = getattr(self.payload, '{field_name}', [])"
                    )
                    lines.append(
                        f"        return [{message_type}.from_proto(item) for item in proto_list] if proto_list else None"
                    )
                else:
                    lines.append(
                        f"        proto_val = getattr(self.payload, '{field_name}', None)"
                    )
                    lines.append(
                        f"        return {message_type}.from_proto(proto_val) if proto_val is not None else None"
                    )
            else:
                lines.append(
                    f"        return getattr(self.payload, '{field_name}', None)"
                )
        else:
            # Scalar or enum fields
            lines.append(f"        return getattr(self.payload, '{field_name}', None)")

    lines.append("")
    lines.append("    @classmethod")
    lines.append(
        "    def from_proto(cls, proto_obj: events_pb2.{0}, **extra):".format(
            proto_name
        )
    )
    lines.append('        """Create event instance from protobuf message."""')
    lines.append("        return cls(payload=proto_obj, **extra)")
    lines.append("")
    lines.append("    def as_dict(self) -> Dict[str, Any]:")
    lines.append('        """Convert protobuf payload to dictionary."""')
    lines.append("        if self.payload is None:")
    lines.append("            return {}")
    lines.append("        return _to_dict(self.payload)")
    lines.append("")
    lines.append("    def __getattr__(self, item: str):")
    lines.append('        """Delegate attribute access to protobuf payload."""')
    lines.append("        if self.payload is not None:")
    lines.append("            return getattr(self.payload, item)")
    lines.append(
        "        raise AttributeError(f\"'{self.__class__.__name__}' object has no attribute '{item}'\")"
    )
    lines.append("")
    return lines


def _render_module_body() -> Tuple[List[str], List[str], List[str]]:
    """Generate message wrappers and event classes.

    Returns:
        Tuple of (message_wrapper_blocks, event_class_blocks, event_class_names)
    """
    # Collect all message types used in events
    message_types = _collect_message_types()

    # Generate message wrapper classes
    message_wrapper_blocks: List[str] = []
    for class_name in sorted(message_types.keys()):
        message_cls = message_types[class_name]
        wrapper_lines = _render_message_wrapper(class_name, message_cls)
        message_wrapper_blocks.append("\n".join(wrapper_lines))

    # Generate event classes
    event_class_blocks: List[str] = []
    event_class_names: List[str] = []

    for proto_name, message_cls in _iter_protobuf_messages():
        class_name = _class_name(proto_name)
        event_class_names.append(class_name)
        class_lines = _render_class(proto_name, message_cls)
        event_class_blocks.append("\n".join(class_lines))

    return message_wrapper_blocks, event_class_blocks, event_class_names


def _build_module() -> str:
    message_wrappers, event_classes, event_names = _render_module_body()

    parts: List[str] = [
        '"""Auto-generated SFU event dataclasses. Do not edit manually."""',
        "# Generated by _generate_sfu_events.py",
        *HEADER_LINES,
    ]

    # Add section header for message wrappers
    if message_wrappers:
        parts.extend(
            [
                "# " + "=" * 78,
                "# Message Type Wrappers",
                "# These are wrappers for protobuf message types used in events",
                "# " + "=" * 78,
                "",
            ]
        )
        parts.extend(message_wrappers)
        parts.append("")

    # Add section header for event classes
    parts.extend(
        [
            "# " + "=" * 78,
            "# Event Classes",
            "# " + "=" * 78,
            "",
        ]
    )
    parts.extend(event_classes)

    # Add exports
    exports_section = [
        "",
        "__all__ = (",
        *[f'    "{name}",' for name in event_names],
        ")",
    ]

    parts.extend(exports_section)

    return "\n".join(parts) + "\n"


def verify_generated_classes() -> bool:
    """Verify that generated classes match protobuf definitions.

    Returns:
        True if all checks pass, False otherwise.
    """
    import importlib.util
    import sys

    # Import the generated module
    target_path = pathlib.Path(__file__).parent / "core" / "edge" / "sfu_events.py"
    if not target_path.exists():
        print("Error: sfu_events.py not found. Run generation first.")
        return False

    # Dynamically load the module
    spec = importlib.util.spec_from_file_location("sfu_events", target_path)
    if spec is None or spec.loader is None:
        print("Error: Could not load sfu_events module")
        return False

    sfu_events = importlib.util.module_from_spec(spec)
    sys.modules["sfu_events"] = sfu_events
    spec.loader.exec_module(sfu_events)

    all_valid = True

    for proto_name, message_cls in _iter_protobuf_messages():
        class_name = _class_name(proto_name)

        # Check if class exists in generated module
        if not hasattr(sfu_events, class_name):
            print(f"✗ Class {class_name} not found in generated module")
            all_valid = False
            continue

        event_class = getattr(sfu_events, class_name)

        # Verify it's a BaseEvent subclass
        if not hasattr(event_class, "__mro__"):
            print(f"✗ {class_name} is not a class")
            all_valid = False
            continue

        # Check field correspondence
        proto_fields = {f.name: f for f in message_cls.DESCRIPTOR.fields}

        # Check that all protobuf fields are accessible via properties
        for field_name, field_desc in proto_fields.items():
            if field_name in {
                "type",
                "event_id",
                "timestamp",
                "session_id",
                "user_metadata",
            }:
                continue  # Skip BaseEvent fields

            if not hasattr(event_class, field_name):
                print(
                    f"✗ {class_name} missing property for protobuf field: {field_name}"
                )
                all_valid = False
                continue

            # Verify it's a property (check on the class itself, not an instance)
            attr = getattr(event_class, field_name, None)
            if not isinstance(attr, property):
                print(
                    f"✗ {class_name}.{field_name} is not a property (type: {type(attr).__name__})"
                )
                all_valid = False
                continue

        print(f"✓ {class_name} verified ({len(proto_fields)} protobuf fields)")

    return all_valid


def verify_field_types() -> None:
    """Verify and display field type mappings for all protobuf messages."""
    print("\n" + "=" * 80)
    print("Field Type Verification Report")
    print("=" * 80 + "\n")

    for proto_name, message_cls in _iter_protobuf_messages():
        class_name = _class_name(proto_name)
        print(f"\n{class_name} ({proto_name}):")
        print(f"  Protobuf type: {message_cls.DESCRIPTOR.full_name}")

        field_descriptors = message_cls.DESCRIPTOR.fields
        if not field_descriptors:
            print("  (no fields)")
            continue

        for field_desc in field_descriptors:
            field_name = field_desc.name
            if field_name in {
                "type",
                "event_id",
                "timestamp",
                "session_id",
                "user_metadata",
            }:
                continue

            python_type = _get_python_type_from_protobuf_field(field_desc)
            proto_type_name = field_desc.type
            label = (
                "repeated"
                if field_desc.is_repeated
                else "optional"
                if hasattr(field_desc, "is_optional")
                else "required"
            )

            print(f"  - {field_name}: type={proto_type_name} ({label}) → {python_type}")


def main() -> None:
    import sys

    # Generate sfu_events.py in the core/edge directory
    target_path = pathlib.Path(__file__).parent / "core" / "edge" / "sfu_events.py"
    target_path.write_text(_build_module(), encoding="utf-8")
    print(f"Regenerated {target_path}")

    # Verify field types
    if "--verify-types" in sys.argv:
        verify_field_types()

    # Verify generated classes
    if "--verify" in sys.argv:
        print("\nVerifying generated classes...")
        if verify_generated_classes():
            print("\n✓ All verifications passed!")
        else:
            print("\n✗ Some verifications failed!")
            sys.exit(1)


if __name__ == "__main__":
    main()
